/*! \file equipstreambuf.hh
    libfsf, generating code to fill structures from configuration files.
    Copyright (C) 2007 Matthieu Nottale. 

    This file is part of libfsfc.

    libfsfc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 3,
    as published by the Free Software Foundation.

    libfsfc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libfsfc.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EQUIPSTREAMBUF_HH
#define EQUIPSTREAMBUF_HH

#include <iostream>
#include <string>

using std::ios;
/// Element of a chain of streambuf
class ChainedStreambuf: public std::streambuf
{
public:
  ChainedStreambuf(std::streambuf & base):base(&base)
  {
    setEmpty();
  }
  virtual ~ChainedStreambuf()
  {
    delete base;
  }
  /// Locate a specific type in chain, return it if found, create it otherwise
  template<class T> static T* getStreambufByType(std::ios& s, T*,
  bool createIfNotFound=true);
  /// Remove a type from the chain if it exists, return true if fould
  template<class T> static bool removeStreambufByType(std::ios& s, T*);
  /// Remvoe all ChainedStreambuf from stream
  static inline void clearChain(std::ios &s);
  /// Returns true if no ChainedStreambuf equips the stream
  static inline bool isChainEmpty(std::ios &s);
protected:
  /// setup buffer with one char
  void setAvailable(char c, bool roomForPutBack = true)
  {
    buf[1] = c;
    setg(roomForPutBack?buf:buf+2, buf+1, buf+2);
  }
  /// setup empty buffers
  void setEmpty(bool roomForPutBack = true)
  {
    setg(roomForPutBack?buf:buf+2, buf+2, buf+2);
  }
  /// Override one so that dynamic_cast may work
  virtual int underflow() { return std::streambuf::underflow();}
  /// Get one char from underlying streambuf
  int baseGet() {return base?base->sbumpc():EOF;}
  /// Remove us from chain, returns base
  virtual std::streambuf* unHook()
  {
    std::streambuf * r = base;
    base = 0;
    return r;
  }
  /// Change our base
  virtual void rebase(std::streambuf* nBase)
  {
    base = nBase;
  }
private:
  std::streambuf * base;
  char buf[2];
};



inline bool ChainedStreambuf::isChainEmpty(std::ios& s)
{
  return !dynamic_cast<ChainedStreambuf*>(s.rdbuf());
}
inline void ChainedStreambuf::clearChain(std::ios& s)
{
  std::streambuf *sb = s.rdbuf();
  ChainedStreambuf* prev = 0;
  while (ChainedStreambuf* csb = dynamic_cast<ChainedStreambuf*>(sb))
  {
    prev = csb;
    sb = csb->base;
  }
  if (!prev)
    return;
  std::streambuf* keep = sb;
  prev->unHook();
  delete s.rdbuf();
  s.rdbuf(keep);
}

template<class T> T* 
ChainedStreambuf::getStreambufByType(std::ios &s,  T*, 
  bool createIfNotFound)
{
  std::streambuf *sb = s.rdbuf();
  while (ChainedStreambuf * csb = dynamic_cast<ChainedStreambuf*>(sb))
  {
    T * target = dynamic_cast<T*>(csb);
    if (target)
      return target;
    else 
      sb = csb->base;
  }
  //not found
  if (!createIfNotFound)
    return 0;
  T * target = new T(*s.rdbuf());
  s.rdbuf(target);
  return target;
}

template<class T> bool
ChainedStreambuf::removeStreambufByType(std::ios& s,  T*)
{
  std::streambuf* sb = s.rdbuf();
  ChainedStreambuf* prev = 0;
  while (ChainedStreambuf* csb = dynamic_cast<ChainedStreambuf*>(sb))
  {
    T * target = dynamic_cast<T*>(csb);
    if (target)
    {
      if (!prev)
	s.rdbuf(target->unHook());
      else
	prev->rebase(target->unHook());
      delete target;
      return true;
    }
    
    prev = csb;
    sb = csb->base;
  }
  //not found
  return false;
}
#endif

