/*! \file fsfc_internal.hh
    libfsf, generating code to fill structures from configuration files.
    Copyright (C) 2007 Matthieu Nottale. 

    This file is part of libfsfc.

    libfsfc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 3,
    as published by the Free Software Foundation.

    libfsfc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libfsfc.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef FSFC_INTERNAL_HH
#define FSFC_INTERNAL_HH

/* Fill Struct From String internal main internal header

*/

#include <ctype.h>
#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <sstream>
#include <fstream>
#include <boost/foreach.hpp>
#include <boost/preprocessor.hpp>
#include <boost/scoped_ptr.hpp>
#include <fsfc/trace.hh>
#include <fsfc/hash.hh>



namespace fsfc 
{

#if defined FSFC_DEFAULTS_TO_SHIFT_OPERATOR \
 && defined FSFC_USE_DYNAMIC_RESOLUTION
#warning "dynamic resolution is useless when defaulting to shift operators"
#undef FSFC_USE_DYNAMIC_RESOLUTION
#endif

struct Marker;
class ParseException: public std::exception
{
public:
  ParseException(std::pair<int ,int> pos, const std::string&msg)
  :line(pos.first), column(pos.second)
  {
    std::ostringstream ss;
    ss<<"error:"<<line<<':'<<column<<':'<<msg;
    message = ss.str();
  }
  virtual const char * what() const throw() 
  {
    return message.c_str();
  }
  virtual ~ParseException() throw() {}
  
private:
  std::string message;
  int line;
  int column;
};

/// Tri-stated option
enum OptionState
{
  /// Option can be set or not
  ALLOW = 0,
  /// Option must not be set
  DENY,
  /// Option must be set
  REQUIRE,
  /// Internal use
  UNSET
};
  
/// Main parser option structure. Can be attached to a field, or a type.
class ParseOptions
{
public:
  /// Ignore indentation
  bool ignoreIndent;
  /// Is the field optional
  bool optional;
  /// Are multiple occurrences of the same field permitted
  bool multipleWrite;
  /// Presence of the fields names (opposed to the ordered list of values
  OptionState fieldName;
  /// Presence of an expected terminator such as '}'
  OptionState terminator;
  /// The structure is on a single line
  OptionState singleLine;
  /// A colon separates the field name from the value
  OptionState colonAfterFieldName;
  /// An equal sign separates the field name from the value
  OptionState equalAfterFieldName;
  /// The field name is written as '<fieldName>'
  OptionState xmlFieldName;
  /// A dash '-' prefixes all container entries (like in yaml)
  OptionState dashInList;
  /// Ctor. Sets reasonable defaults
  ParseOptions():
   ignoreIndent(false), optional(false), multipleWrite(false),
  fieldName(ALLOW), terminator(ALLOW), singleLine(ALLOW),
  colonAfterFieldName(ALLOW), equalAfterFieldName(ALLOW),
  xmlFieldName(ALLOW), dashInList(ALLOW), allowNoise(false) {}
  /// Fill struct from configuration
  void parse(const std::string & str);
  /// Merge from an other configuration
  void merge(const ParseOptions & po);
  /// Crawl around for anything that might be a token. Usage discouraged.
  bool allowNoise;
private:

};
std::ostream& operator << (std::ostream&k, const ParseOptions&);

#define PARSE_OPTION_OPTIONSTATE_LIST BOOST_PP_TUPLE_TO_LIST(7, \
  (fieldName, terminator, singleLine, colonAfterFieldName, equalAfterFieldName,\
  xmlFieldName, dashInList))
#define PARSE_OPTION_BOOL_LIST BOOST_PP_TUPLE_TO_LIST(3, \
  (optional, multipleWrite, ignoreIndent))

/// MetaStructure interface describing a class
class StructDesc
{
public:
  /// Is givent string a field name?
  virtual bool isField(const std::string &) {return false;}
  /// Can the field be written to?
  virtual bool canWriteField(const std::string &) {return true;}
  /// Fill struct content from is
  virtual void fillStruct(std::istream & is, Marker &marker);
  /// Overload with empty marker
  void fillStruct(std::istream &is);
  /// Is struct correctly filled in? Return "" if yes, an error message if no.
  virtual std::string validate() { return std::string();}
  /// Get StructDesc for a member
  virtual StructDesc * getDescFor(const std::string &)  { return 0;}
  /// Get next field name
  virtual const std::string & nextField() { static std::string e(""); return e;}
  /// Get options for given field
  virtual const ParseOptions & getOptions(const std::string &) = 0;
  /// Get options for class
  virtual const ParseOptions & getClassOptions() 
    {static ParseOptions o; return o;}
  /// Fill os with struct content, return true if last outputed char was "\n"
  virtual bool outputStruct(std::ostream &o, const ParseOptions &p, int indent);
  /// Overload with default parameters
  void outputStruct(std::ostream &o);
  /// Get current type name (for debugging)
  virtual std::string getName() {return "";}
  
  virtual ~StructDesc() {}
protected:
  /// Throw a parse exception, containing the line/column of current stream pos.
  void throwException(std::istream &is, const std::string &msg);
};



/// Trait used to store parsing options per class
template<class T> class DescOptTrait
{
public:
  static const bool multipleWrite = false;
  static const OptionState fieldName =  ALLOW;
  static const OptionState terminator = ALLOW;
  static const OptionState colonAfterFieldName = ALLOW;
  static const OptionState equalAfterFieldName = ALLOW;
  static const OptionState xmlFieldName = ALLOW;
  static const OptionState singleLine = ALLOW;
};

#define ONELINE_TRAIT(T) \
template<> class DescOptTrait<T>              \
{                                             \
public:                                       \
  static const bool multipleWrite = false;    \
  static const OptionState fieldName =  DENY; \
  static const OptionState terminator = DENY; \
  static const OptionState colonAfterFieldName = DENY;\
  static const OptionState equalAfterFieldName = DENY;\
  static const OptionState xmlFieldName = DENY;       \
  static const OptionState singleLine = REQUIRE;      \
};

ONELINE_TRAIT(int)
ONELINE_TRAIT(unsigned int)
ONELINE_TRAIT(float)
ONELINE_TRAIT(double)
ONELINE_TRAIT(std::string)
/// Fill option struct from a trait struct
template<class T> void fillOptions(ParseOptions &p, T t)
{
  p.multipleWrite = t.multipleWrite;
  p.fieldName = t.fieldName;
  p.terminator = t.terminator;
  p.colonAfterFieldName = t.colonAfterFieldName;
  p.equalAfterFieldName = t.equalAfterFieldName;
  p.xmlFieldName = t.xmlFieldName;
  p.singleLine = t.singleLine;
}

bool gotoNextToken(std::istream & is, Marker& marker,  std::string& eaten_,
  const ParseOptions&);

//wrapper to "untemplate" fillStruct
class BaseBaseStructDesc: public StructDesc
{
public:
  virtual void callShiftOp(std::istream& )=0;
  virtual void fillStruct(std::istream& is, Marker& m);
};
/// StructDesc implementation using the << operator
template <typename T> class BaseStructDesc: public BaseBaseStructDesc
{
public:
  T & ptr;
  void callShiftOp(std::istream& is)
  {
    is >> ptr;
    if (!is.good())
      throwException(is, "error parsing basic type");
  }
  BaseStructDesc(void * ptr_) :ptr( *(T*)ptr_) {}
 
  bool outputStruct(std::ostream &o, const ParseOptions &, int )
  {
    o << ptr;
    return false;
  }
  static StructDesc* createInstance(void * p) {return new BaseStructDesc<T>(p);}
  virtual const ParseOptions & getOptions(const std::string &)
  {
    static ParseOptions opt;
    fillOptions(opt, DescOptTrait<T>());
    return opt;
  }
  virtual const ParseOptions & getClassOptions()
  {
    return getOptions("");
  }
  virtual std::string getName() {return "basic";}
};


#if !defined FSFC_DEFAULTS_TO_SHIFT_OPERATOR
/// Unknown Desc class throwing an error when constructed
class UnknownDescClass: public StructDesc
{
public:
  UnknownDescClass(void*) {throw std::string ("Unknown metaclass");};
};
#endif

/// Trait used to access the StructDesc class
template<class T> class DescTrait 
{
public:
  #ifdef FSFC_DEFAULTS_TO_SHIFT_OPERATOR
  typedef BaseStructDesc<T> type;
  #else
  typedef UnknownDescClass type;
  #endif
};

/// Register an association between a type and its StructDesc type
#define REGISTER_DESCTRAIT(t, mt) \
template<> class DescTrait<t> {public: typedef mt type;}






typedef StructDesc *(*DescFactory)(void *);

#if defined FSFC_USE_DYNAMIC_RESOLUTION
class Registrant {
public:
  Registrant(const std::string &name, DescFactory ptr);
};


/// Register a class with its DescStruct trait and its DescStruct factory
#define REGISTER_DESC(classname, descclassname) \
  Registrant _r_##classname(#classname, &descclassname::createInstance);\
  REGISTER_DESCTRAIT(classname, descclassname)
/// Register an alternate name with a Deststruct Factory
#define REGISTER_NAME(name, uid, cls) \
  Registrant _r_##uid(#name, &cls::createInstance);
#else
 #define REGISTER_DESC REGISTER_DESCTRAIT
 #define REGISTER_NAME(a,b,c)
#endif
  




//! Base class used by custom autogenerated implementations.
/*! This class holds the list of field names, a hash of ParseOptions, and a
    hash of bools to remember which field was written to.
*/
class StructDescImpl: public StructDesc {
public:
  StructDescImpl() : _meta_pos(0) {}
  virtual ~StructDescImpl() {}
  libport::hash_map<std::string, bool> written;
  libport::hash_map<std::string, ParseOptions> options;
  unsigned int _meta_pos;
  std::vector<std::string> fields; 
  
  const std::string & nextField() 
  {
    static std::string e;
    ++_meta_pos; 
    if (_meta_pos-1 < fields.size()) 
      return fields[_meta_pos-1]; 
    else 
      return e;
  }
  bool isField(const std::string & name)
  {
    return ::libport::mhas(written, name);
  }
  bool canWriteField(const std::string & name)
  {
    return ::libport::mhas(written, name) && 
    ( !written[name] || getOptions(name).multipleWrite);
  }
  std::string validate()
  {
    typedef std::pair<const std::string, bool> SBPair;
    BOOST_FOREACH(SBPair &item, written)
    if (!item.second && !options[item.first].optional)
      return "struct field "+item.first+" not set";
    return std::string();
  }
 
  
  virtual const ParseOptions & getOptions(const std::string& name)
  {
    return options[name];
  }
};

//try first dynamic, then static resolution
#define CREATEDESC(cls, param) fsfc::createDesc((cls*)0, #cls, (void *)param)

#if defined FSFC_USE_DYNAMIC_RESOLUTION
 /// Singleton smart pointer that creates the object on demand.
  template<class T>
  class SingletonPtr
  {
  public:
    operator T* ()
    {
      return instance();
    }
    operator T& ()
    {
      return *instance();
    }

    T* operator ->()
    {
      return instance();
    }

  private:
    T* instance()
    {
      static T* ptr = 0;
      if (!ptr)
        ptr = new T();
      return ptr;
    }
  };

extern SingletonPtr<libport::hash_map<std::string, DescFactory> >
         descList;
#endif


/* I would have used boost, but boost.org was offline *again* when I needed the
 trait. */
template<class T> class noconst
{
public:
  typedef T type;
};
template<class T> class noconst<const T>
{
public:
  typedef T type;
};

template<class T> inline StructDesc * createDesc(T *, const std::string& name,
  void *p)
{
  TRACE(100, "createDesc call "<<name<<" "  <<' '<<p);
#if defined FSFC_USE_DYNAMIC_RESOLUTION
  if (libport::mhas(*descList, name)) 
  {
    TRACE(100, "  got result from descList");
    return (*descList)[name](p);
  }
  else 
#endif
  {
    TRACE(100, "  no descL result, trying static...");
    return new typename DescTrait<typename noconst<T>::type>::type(p);
  }
}



class BaseContainerDesc: public StructDesc
{
public:
  virtual void fillStruct(std::istream & is, Marker &marker);
};


template<class C> class ContainerDesc: public BaseContainerDesc
{
public:
  ContainerDesc(void * p):container(*(C*)p) {
    TRACE(4, "create containerdesc from "<<p);
  }
  
  virtual const std::string& nextField() 
  {
    TRACE_ME;
    static std::string s(" _"); return s;
  }
  virtual StructDesc * getDescFor(const std::string &) 
  {
    TRACE_ME;
    //hack: we now exactly one call to getDescFor() happens before a parse
    //attepmt, so push_back a new element to the list here
    typename C::value_type val = typename C::value_type();
    container.push_back(val);
    return createDesc((typename C::value_type*)0, " _", &container.back());
  }
  virtual std::string getName() {return "ContainerDesc";}
  C &container;
  virtual ParseOptions & getOptions(const std::string&)
  {
    static ParseOptions opt;
    fillOptions(opt, DescOptTrait<typename C::value_type>());
    return opt;
  }
  
  virtual bool outputStruct(std::ostream &o, const ParseOptions & mo, 
  int indent)
  {
    TRACE_ME;
    std::string s;
    s.insert(s.begin(), indent+2, ' ');
    typedef typename C::value_type cval;
    BOOST_FOREACH(cval &v, container)
    {
      boost::scoped_ptr<StructDesc> sd (createDesc((typename C::value_type*)0,
	" _", &v));
      ParseOptions po  = sd->getClassOptions();
      TRACE(5, "mo = "<<mo);
      TRACE(5, "po = "<<mo);
      if (mo.singleLine != REQUIRE && po.singleLine == REQUIRE)
	o << s;
      if (mo.dashInList == REQUIRE)
	o << "- ";
      sd->outputStruct(o, po, indent+2);
      o << '\n';
    }
    return !container.empty();
  }
};

template<typename T> class DescTrait<std::vector<T> >
{
public:
  typedef ContainerDesc<std::vector<T> > type;
};
template<typename T> class DescTrait<std::list<T> >
{
public:
  typedef ContainerDesc<std::list<T> > type;
};

template<typename T> class DescOptTrait<std::vector<T> >
{
public:
  static const bool multipleWrite = true;
  static const OptionState fieldName =  ALLOW;
  static const OptionState terminator = ALLOW;
  static const OptionState colonAfterFieldName = ALLOW;
  static const OptionState equalAfterFieldName = ALLOW;
  static const OptionState xmlFieldName = ALLOW;
  static const OptionState singleLine = ALLOW;
};
template<typename T> class DescOptTrait<std::list<T> >
{
public:
  static const bool multipleWrite = true;
  static const OptionState fieldName =  ALLOW;
  static const OptionState terminator = ALLOW;
  static const OptionState colonAfterFieldName = ALLOW;
  static const OptionState equalAfterFieldName = ALLOW;
  static const OptionState xmlFieldName = ALLOW;
  static const OptionState singleLine = ALLOW;
};


class MetaString: public BaseStructDesc<std::string>
{
public:
  MetaString(void * p): BaseStructDesc<std::string>(p) {}
  static StructDesc* createInstance(void * p) {return new MetaString(p);}
  virtual std::string getName() {return "MetaString";}
  virtual void fillStruct(std::istream &is, Marker &);
};
REGISTER_DESCTRAIT(std::string, MetaString);

/// operator >> implementation
#define RIGHTSHIFT_OPERATOR(cls) \
std::istream & operator >> (std::istream &is, cls& o) { \
  TRACE_ME; \
  fsfc::parse(o, is); \
  return is;\
}

#ifndef LEFTSHIFT_OPERATOR
#define LEFTSHIFT_OPERATOR(cls) \
std::ostream& operator <<(std::ostream& os, const cls &c) {\
  fsfc::output(c, os);\
  return os;\
}
#endif

#ifndef SHIFT_OPERATORS
#define SHIFT_OPERATORS(cls) \
   LEFTSHIFT_OPERATOR(cls) \
   RIGHTSHIFT_OPERATOR(cls)
#endif

// Here we define the macros exported by xmltoclassdesc

//Once per class: start the <class>Decl, inheriting StrucDescImpl
#define HANDLE_CLASS_BEGIN(name, m)  \
class m##Desc: public fsfc::StructDescImpl {\
public: \
  name & ref; \
  std::string myName; 

#define HANDLE_CLASS_COMMENT(ca, cb) \
static const char * getComment(const char * parm = #ca #cb) {return parm;} \
static bool parsedComment; \
static fsfc::ParseOptions  classOptions;

#define HANDLE_CLASS_BEGIN_BIND(name, m) \
  const fsfc::ParseOptions & getClassOptions() { return classOptions; } \
  static fsfc::StructDesc* createInstance(void* p) {return new m##Desc(p);} \
  m##Desc(void * ref_):ref(*(name*)ref_), \
    myName(#name) { \
    TRACE_ME; \
    if (!parsedComment) { \
      parsedComment = true; classOptions.parse(getComment());}
    
    //in constructor, initialise written hash to false for all fields, and
    //create list of field names
    //also initialise options
    #define UTryBindVar(className, fieldName,inu) \
      written[#fieldName] = false; \
      fields.push_back(#fieldName); \
      fillOptions(options[#fieldName], fsfc::DescOptTrait<className>()); \
      StructDesc * sd_##fieldName = CREATEDESC(className, 0); \
      options[#fieldName].merge(sd_##fieldName->getClassOptions()); \
      delete sd_##fieldName;
      #define HANDLE_VAR_COMMENT(cn, fieldName, ca, cb) \
      {\
	fsfc::ParseOptions po;\
	po.parse(#ca #cb);\
	options[#fieldName].merge(po); \
      }
  #define HANDLE_CLASS_END_BIND(nm)  } \
  fsfc::StructDesc * getDescFor(const std::string &name) { \
    (void)name;
    
    //if field passed to getDescFor exists, return its Desc class
    //by calling CREATEDESC
    #define UTryBindVarImpl(className, fieldName, inu)\
    TRACE(100, myName <<":scanning "<<name<<" == "<<#fieldName << ' '<<#className ); \
    if (name == #fieldName){\
      TRACE(11, "ok, createMeta("<<#className<<","<<&ref.fieldName<<')'); \
      fsfc::StructDesc * m = CREATEDESC(className, &ref.fieldName); if (m) written[name] = true;return m; }
    
  #define HANDLE_CLASS_END(className, cpn) return 0;}\
  std::string getName() {return myName;} \
}; \
fsfc::ParseOptions cpn##Desc::classOptions; \
bool cpn##Desc::parsedComment = false; \
REGISTER_DESC(className, cpn##Desc); \
SHIFT_OPERATORS(className);



//useless macros exported by xmltoclassdesc.pl
#define HANDLE_CLASS_BEGIN_CTORS(a, b)
#define HANDLE_CLASS_END_CTORS(a)
#define UTryBind(a,b)
#define InheritBind(a,b,c,d)


#if !defined FSFC_DEFAULTS_TO_SHIFT_OPERATOR
// Those are useless if above flag is set, as operator >> is used as fallback.
REGISTER_DESC(int, BaseStructDesc<int>);
REGISTER_DESC(double, BaseStructDesc<double>);
REGISTER_DESC(float, BaseStructDesc<float>);
REGISTER_DESC(long, BaseStructDesc<long>);
REGISTER_DESC(short, BaseStructDesc<short>);
#endif


template<class T> void parse(T& s, std::istream& is)
{
  boost::scoped_ptr<StructDesc> m (CREATEDESC(T,&s)); 
  m->fillStruct(is);
}
template<class T> void output(T& s, std::ostream& os)
{
  boost::scoped_ptr<StructDesc> m (CREATEDESC(T,&s));
  m->outputStruct(os);
}

template<class T> void parse(T& s, const std::string& fname)
{
  std::ifstream ifs;
  ifs.open(fname.c_str(), std::ios::in);
  parse(s, ifs);
  ifs.close();
}
template<class T> void output(T& s, std::string& fname)
{
  std::ofstream ofs;
  ofs.open(fname.c_str(), std::ios::out);
  output(s, ofs);
  ofs.close();
}

}

/*
new version with the namespace. no work-in-all-cases version seems possible
fsfc:: or ::fsfc:: does not work: it gives
error: extra qualification ignored
or 
error: global qualification of class name is invalid before '{' token
*/
#undef REGISTER_DESCTRAIT
#define REGISTER_DESCTRAIT(t, mt) \
namespace fsfc { template<> class DescTrait<t> {public: typedef mt type;};}

#endif
