/*! \file fsfc.cc
    libfsf, generating code to fill structures from configuration files.
    Copyright (C) 2007 Matthieu Nottale. 

    This file is part of libfsfc.

    libfsfc is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 3,
    as published by the Free Software Foundation.

    libfsfc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libfsfc.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cassert>
#include <ctype.h>
#include <boost/scoped_ptr.hpp>
#include <fsfc/undoablestream.hh>
#include <fsfc/poscountingstream.hh>
#include <fsfc/commentstripstream.hh>
#include "fsfc/fsfc.hh"

namespace fsfc
{

struct Marker 
{
  Marker():newLine(false), enforceNewLine(true), spaces(0), enforceIndent(true)
  {}
  std::string terminator;
  bool newLine;
  /// if set to false, determine new line instead of using it
  bool enforceNewLine; 
  int spaces;
  bool enforceIndent;
  void parse(std::istream & is);
  bool conflicts(const Marker &b);
};
bool gotoNextToken(std::istream & is, 
  Marker & marker,
  std::string &, 
  const ParseOptions &po);

#if defined FSFC_USE_DYNAMIC_RESOLUTION
Registrant::Registrant(const std::string &name, DescFactory ptr)
{
  (*descList)[name] = ptr;
}
SingletonPtr<libport::hash_map<std::string, DescFactory> > descList;
#endif

inline std::ostream & operator << (std::ostream &o, const Marker&m)
{
  o <<'[';
  if (m.enforceNewLine)
    o << m.newLine;
  else
    o << '?';
  return o << "," << m.spaces << ",'" << m.terminator << "']";
}

inline std::pair<int, int> getCorrectedPosition(std::istream &is)
{
  std::pair<int,int> p = getStreamPosition(is);
  std::string s = undoBuffer(is);
  TRACE(100, " raw pos "<<p.first<<' ' <<p.second<<" buf '"<<s<<"'");
  int nline = 0;
  for (unsigned int i=0; i < s.length(); i++)
    if (s[i] == '\n')
    nline++;
  if (nline)
  {
    p.first -= nline;
    p.second = getNColsForLine(is, p.first) - s.find_first_of('\n');
  }
  else
    p.second -= s.length();
  TRACE(100, "corrected pos "<<p.first<<' ' <<p.second);
  return p;
}
#define PARSE_ERROR(stream, msg) do{\
  std::stringstream ss;             \
  ss << msg;                        \
  throw ParseException(getCorrectedPosition(is), ss.str()); }\
  while(0)
void StructDesc::throwException(std::istream &is, const std::string &msg)
{
  PARSE_ERROR(is, msg);
}
void StructDesc::fillStruct(std::istream & is)
{
  Marker ma;
  ma.enforceNewLine = false;
  ma.enforceIndent = false;
  fillStruct(is, ma);
}

enum ConfigMode 
{
  MODE_UNKNOWN,
  MODE_KEYVALUE,
  MODE_VALUELIST,
  MODE_NOISE
};

std::ostream & operator <<(std::ostream &s, const ParseOptions &po)
{
  return s << "opt="<<po.optional<<" mw="<<po.multipleWrite
    <<" fn="<<po.fieldName<<" ter="<<po.terminator
    <<" sl="<<po.singleLine<<" col="<<po.colonAfterFieldName
    <<" equ="<<po.equalAfterFieldName<<" xml="<<po.xmlFieldName
    <<" igi="<<po.ignoreIndent;
}


/// Try to eat a character, ignoring whitespaces, restauring in case of failure
bool tryToEat(std::istream &is, char c)
{
  std::string eaten;
  while (is.good())
  {
    char ch = is.get();
    if (ch == c)
      return true;
    if (!isspace(ch))
    {
      undoStream(is, eaten+ch);
      return false;
    }
    eaten += ch;
  }
  undoStream(is, eaten);
  return false;
}
class Token: public std::string
{
public:
  void operator= (const std::string& b)
  {
    this->std::string::operator=(b);
  }
};

std::istream& operator >> (std::istream&is, Token& t)
{
  t.clear();
  while (is.good())
  {
    int c = is.get();
    if (isalpha(c) 
      || c == '_'
      || c == '>'
      || (!t.empty() && isdigit(c))
      || (t.empty() && c == '<'))
      t += c;
    else
    {
      undoStream(is, c);
      break;
    }
  }
  return is;
}

class Noise: public std::string
{
public:
};
std::istream& operator >> (std::istream&is, Noise& t)
{
  t.clear();
  while (is.good())
  {
    int c = is.get();
    if (isalpha(c) || c== '<')
    {
      undoStream(is, c);
      break;
    }
    t += c;
  }
  return is;
}
/** Try hard to obtain a field, check parse constraints, return it
\param eaten will contain what was eaten in stream, in case of success only
*/
std::string getField(StructDesc & sd, const ParseOptions & po, 
  std::istream &is,  std::string & eaten_, std::string& childTerm)
{
  TRACE_ME;
  std::string::size_type pos;
  bool xml=false, colon=false, equal=false;
  Token f;
  is >> f;
  std::string eaten;
  eaten = f;
  bool gotField = false;
  if (sd.isField(f))
    gotField = true;
  TRACE(10, f <<' '<<gotField<<" so far");
  if (!gotField && po.colonAfterFieldName != DENY 
    && (pos=f.find_first_of(':')) != f.npos)
  {
    std::string pf = f.substr(0, pos);
    if (sd.isField(pf))
    {
      colon = true;
      gotField = true;
      undoStream(is, f.substr(pos+1, f.npos));
      f = pf;
      eaten = f;
    }
  }
  if (!gotField && po.equalAfterFieldName != DENY 
    &&  (pos=f.find_first_of('=')) != f.npos)
  {
    std::string pf = f.substr(0, pos);
    if (sd.isField(pf))
    {
      equal = true;
      gotField = true;
      undoStream(is, f.substr(pos+1, f.npos));
      f = pf;
      eaten = f;
    }
  }
  if (!gotField && po.xmlFieldName != DENY && f.length() >2
    && f[0] == '<' && (pos = f.find_first_of('>')) != f.npos)
  {
    std::string pf = f.substr(1, pos-1);
    if (sd.isField(pf))
    {
      TRACE(10, "take " << pf << " undo "<<f.substr(pos+1, f.npos));
      undoStream(is, f.substr(pos+1, f.npos));
      xml = true;
      gotField = true;
      f = pf;
    }
  }
  
  if (!gotField)
  {
    if (po.allowNoise)
      return f;
    undoStream(is, eaten);
    return std::string();
  }
  
  //now check field-value separator and field type
  if (!xml && po.xmlFieldName == REQUIRE)
  {
    TRACE(6, "field "<<f<<" found but not in xml");
    undoStream(is, eaten);
    return std::string();
  }
  if (! equal && po.equalAfterFieldName != DENY && tryToEat(is, '='))
    equal = true;
  if (! colon && po.colonAfterFieldName != DENY && tryToEat(is, ':'))
    colon = true;
  
  if (xml)
    childTerm = "</" + f +'>';
  if ( (!equal && po.equalAfterFieldName == REQUIRE)
    || (!colon && po.colonAfterFieldName == REQUIRE))
  {
    TRACE(6, "expected separator not found");
    undoStream(is, eaten);
    childTerm.clear();
    return std::string();
  }
  eaten_ += eaten;
  return f;
}
class Unequipstream
{
public:
  Unequipstream(std::ios& s) :s(s) {}
  ~Unequipstream()
  {
    ChainedStreambuf::clearChain(s);
  }
  std::ios& s;
};
void BaseContainerDesc::fillStruct(std::istream & is, Marker &marker)
  {
    TRACE_ME;
    boost::scoped_ptr<Unequipstream> unEquip;
    if (ChainedStreambuf::isChainEmpty(is))
    {
      // Ensure all our modifiers are present. Order is important
      ChainedStreambuf::getStreambufByType(is, (PosCountingStreambuf*)0, true);
      ChainedStreambuf::getStreambufByType(is, (CommentStripStreambuf*)0, true);
      ChainedStreambuf::getStreambufByType(is, (UndoableStreambuf*)0, true);
      unEquip.reset(new Unequipstream(is)); 
    }
    TRACE(3, "container fillStruct for "<<getName());
    ParseOptions parseOpts = getClassOptions();
    TRACE(5, " opts: "<<parseOpts);
    std::pair<int, int> p = getCorrectedPosition(is);
    TRACE(10, "position "<<p.first<<' '<<p.second);
    std::string eaten;
    /* This is tricky, as the list in itself might have no token. So we need to
    propagate our setup to the children if no terminator was used.
    */
    OptionState dash = UNSET;
    if (parseOpts.dashInList == DENY)
      dash = DENY;
    if (parseOpts.dashInList == REQUIRE)
      dash = REQUIRE;
    do
    {
      if (dash != DENY)
      {
	int c1 = is.get();
	int c2 = is.get();
	if (!is.good())
	  return;
	bool havedash = (c1 == '-' && isspace(c2));
	if (!havedash)
	{
	  undoStream(is, c2);
	  undoStream(is, c1);
	}
	else if (c2 == '\n')
	  undoStream(is, c2); //we need to restore this, if space we dont care
	  
	if (!havedash && dash == REQUIRE)
	  PARSE_ERROR(is, "expected - for list item not found");
	//Enforce consistency between all the list items.
	if (dash == UNSET)
	{
	  dash = havedash? REQUIRE:DENY;
	  TRACE(10, "dash set to " << dash);
	}
      }
      
      TRACE(10, "marker " << marker);
      Marker childMarker;
      
      childMarker.parse(is);
      TRACE(4, "child marker: "<<childMarker);
      if  (!childMarker.spaces 
	&& !childMarker.newLine 
	&& childMarker.terminator.empty()
	)
      {
	TRACE(5, "unsafe markerless list");
	childMarker.newLine = false;
	/// If we know we are on one line, dont allow child to spread.
	if (!marker.enforceNewLine || marker.newLine)
	  childMarker.enforceNewLine = false;
	childMarker.spaces = marker.spaces;
      }
      TRACE(4, "transformed child marker: "<<childMarker);
      boost::scoped_ptr<StructDesc> childMeta (getDescFor(""));
      TRACE(5, getName() << " got meta "<<childMeta << ' ' 
	<<(childMeta?childMeta->getName():std::string() ) );
      if (!childMeta)
	PARSE_ERROR(is, "unable to handle type for container field");
      p = getCorrectedPosition(is);
      TRACE(10, "position "<<p.first<<' '<<p.second);
      childMeta->fillStruct(is, childMarker);
      TRACE(4, "back to fillStruct "<<getName());
      p = getCorrectedPosition(is);
      TRACE(10, "position "<<p.first<<' '<<p.second);
  }
  while (gotoNextToken(is, marker, eaten, parseOpts));
  
  }
  
void StructDesc::fillStruct(std::istream & is, Marker &marker)
{
  TRACE_ME;
  boost::scoped_ptr<Unequipstream> unEquip;
  if (ChainedStreambuf::isChainEmpty(is))
  {
    // Ensure all our modifiers are present. Order is important
    ChainedStreambuf::getStreambufByType(is, (PosCountingStreambuf*)0, true);
    ChainedStreambuf::getStreambufByType(is, (CommentStripStreambuf*)0, true);
    ChainedStreambuf::getStreambufByType(is, (UndoableStreambuf*)0, true);
    unEquip.reset(new Unequipstream(is)); 
  }
  TRACE(3, "fillStruct for "<<getName());
  
  ParseOptions parseOpts = getClassOptions();
  TRACE(5, " opts: "<<parseOpts);
    
  ConfigMode m = MODE_UNKNOWN;
  if (getClassOptions().allowNoise)
  {
    TRACE(1, "##noise mode");
    m = MODE_NOISE;
  }
  
  if (parseOpts.fieldName == DENY)
    m = MODE_VALUELIST;
  if (parseOpts.fieldName == REQUIRE)
    m = MODE_KEYVALUE;
  
  std::string eaten;
  /*
  If we were called from another fillStruct, we now next character is not a
  space, but the begining of a field name or value.
  But if we were called from operator >> its not true, so check
  */
  int t = is.get();
  undoStream(is, static_cast<char>(t));
  if (isspace(t))
  {
    Marker tm;
    tm.parse(is);
    if (marker.enforceNewLine)
      assert (!tm.newLine);
    if (marker.enforceIndent)
      assert (!tm.spaces);
    if (tm.spaces && !marker.enforceIndent)
    {
      marker.enforceIndent = true;
      marker.spaces = tm.spaces;
      TRACE(10, "detected indent " << tm.spaces);
    }
  }
  else if (!marker.enforceIndent)
  {
    marker.enforceIndent = true;
    marker.spaces = 0;
  }
  do
  {
    if (!is.good())
      break;
    std::pair<int, int> p = getCorrectedPosition(is);
    TRACE(10, "position "<<p.first<<' '<<p.second);
    std::string tok;
    std::string childTerm;
    if (m == MODE_UNKNOWN)
    {
      tok = getField(*this, parseOpts, is, eaten, childTerm);
      TRACE(4, "token '"<<tok<<"'");
      if (!tok.empty())
	m = MODE_KEYVALUE;
      else
	m = MODE_VALUELIST;
    }
    
    
    switch (m)
    {
    case MODE_VALUELIST:
      tok = nextField();
      TRACE(10, "nextField " << tok);
      if (tok.empty())
      {
	undoStream(is, eaten);
	if (!marker.terminator.empty())
	  PARSE_ERROR(is, "missing terminator: " << marker.terminator);
	std::string er = validate();
	if (!er.empty())
	  PARSE_ERROR(is, er);
	return; //XXX eat some markers?
      }
      break;
    case MODE_NOISE:
      tok = getField(*this, parseOpts, is, eaten, childTerm);
      if (tok.empty())
      {
	Noise n;
	is >> n; //because getField eats nothing if not a token
      }
      TRACE(5, "noise token "<<tok);
      if (!isField(tok))
	continue;
      TRACE(5, "good token "<<tok);
      break;
    
    case MODE_KEYVALUE:
      TRACE(6, "getfield? "<< parseOpts);
      if (tok.empty()) //=== true except on first iteration
	tok = getField(*this, parseOpts, is, eaten, childTerm);
      TRACE(6, "field= "<<tok);
      break;
    case MODE_UNKNOWN:
      assert(false);
      break;
    }
   
    TRACE(4, "  token "<<tok);
    if (!canWriteField(tok)) {
      TRACE(2, "cant write to field "<<tok <<" restoring '"<<eaten<<"'");
      undoStream(is, eaten);
      break;
    }

    Marker childMarker;
    childMarker.parse(is);
    TRACE(4, "child marker: "<<childMarker<<" childTerm "<<childTerm);
    if (!childMarker.terminator.empty() && !childTerm.empty())
      PARSE_ERROR(is, "multiple start markers");
    if (!childTerm.empty())
      childMarker.terminator = childTerm;
    TRACE(4,"marker check "<<childMarker << ' ' <<marker);
    if (childMarker.conflicts(marker))
      PARSE_ERROR(is, "incorrect indent");
    boost::scoped_ptr<StructDesc> childMeta (getDescFor(tok));
    TRACE(5, getName() << " got meta "<<childMeta << ' ' 
    <<(childMeta?childMeta->getName():std::string() ) );
    if (!childMeta)
      PARSE_ERROR(is, "unable to handle type for field "<<tok);
    
    eaten.clear();
    p = getCorrectedPosition(is);
    TRACE(10, "position "<<p.first<<' '<<p.second);
    childMeta->fillStruct(is, childMarker);
    TRACE(4, "back to fillStruct "<<getName());
    p = getCorrectedPosition(is);
    TRACE(10, "position "<<p.first<<' '<<p.second);
  }
  while (gotoNextToken(is, marker, eaten, parseOpts));
  TRACE(10, "validating "<<getName());
  std::string err = validate();
  if (!err.empty())
   PARSE_ERROR(is, err);
}

void
BaseBaseStructDesc::fillStruct(std::istream &is, Marker &m) 
{ 
  TRACE_ME;
  (void)m;
  //TRACE(10, "basic fillstruct with");
  callShiftOp(is);
  //TRACE(10, "filled field with '"<<ptr<<"'");
  if (!is.good())
    throwException(is, "error parsing basic type");
  std::string eaten;
  ParseOptions opt;
  //fillOptions(opt, DescOptTrait<T>());
  if (!m.terminator.empty())
    gotoNextToken(is, m, eaten, opt);
}

void
StructDesc::outputStruct(std::ostream &o)
{
  const ParseOptions & po = getClassOptions();
  outputStruct(o, po, 0);
}

bool
StructDesc::outputStruct(std::ostream &o, const ParseOptions &parent, 
  int indent)
{
  TRACE_ME;
  TRACE(4, "outputStruct "<<getName());
  TRACE(5, " parent opts: "<<parent);
  std::string indentStr;
  indentStr.insert(indentStr.begin(), indent, ' ');
  std::string field = nextField();
  bool newline = false;
  while (true)
  {
    TRACE(5, "nextField " << field);
    ParseOptions po = getOptions(field);
    if (parent.singleLine == REQUIRE)
      po.singleLine = REQUIRE;
    if (!(parent.singleLine == REQUIRE))
      o << indentStr;
    if (parent.fieldName != DENY)
    {
      if (parent.xmlFieldName == REQUIRE)
        o << '<' << field << '>';
      else 
	o << field;
      o << ' ';
      if (parent.colonAfterFieldName == REQUIRE)
	o << ": ";
      if (parent.equalAfterFieldName == REQUIRE)
	o << "= ";
    }
    if (po.terminator == REQUIRE)
      o << '{';
    if (po.singleLine != REQUIRE)
      o << '\n';
    boost::scoped_ptr<StructDesc> childDesc (getDescFor(field));
    assert(childDesc);
    TRACE(6, "outputing "<<childDesc->getName()<<" with opt "<<po);
    newline = childDesc->outputStruct(o, po, indent+2);
    
    std::string terminator;
    if (po.terminator == REQUIRE)
      terminator = '}';
    if (parent.xmlFieldName == REQUIRE)
      terminator = " </" + field + '>';
    if (po.singleLine != REQUIRE)
    {
      if (!newline) 
      {	
	TRACE(6,"newline");
	o << '\n';
	newline = true;
      }
      if (!terminator.empty())
	o << indentStr << "  ";
    }
    o << terminator;
    
    field = nextField();
    if (!field.empty())
    {
      //avoid a double newline
      if (!newline)
      {
	if (parent.singleLine != REQUIRE)
	{
	  TRACE(6,"newline");
	  o << '\n';// << indentStr;
	  newline = true;
	}
	else
	  o << ' ';
      }
    }
    else
      break;
  }
    
  TRACE(4, "end outputStruct "<<getName());
  return newline;
}

///checks if marker conflicts with parent marker
bool Marker::conflicts(const Marker &b)
{
  if (!terminator.empty())
    return false;
  if (!newLine)
    return false; //one-line: ok
  if (newLine && !b.newLine && b.enforceNewLine)
    return true;
  return spaces < b.spaces;
}


/** Parse new marker
*/
void Marker::parse(std::istream &is)
{
  TRACE_ME;
  terminator.clear();
  newLine = false;
  spaces = 0;
  while (is.good())
  {
    int a = is.get();
    TRACE(6,"  got "<<a);
    if (isspace(a))
    {
      //ignore whitespaces until we hit a \n
      if (a=='\n') 
      {
	newLine = true;
	spaces = 0;
      }
      else
	spaces++;
      continue;
    }
    if (a == '{' && terminator.empty())
    {
      terminator = "}";
      newLine = false;
      spaces = 0;
      continue;
    }
    undoStream(is, a);
    if (!newLine)
      spaces = 0;
    TRACE(6,"  returning");
    return;
  }
}


/** Advance stream until start of new field, return false if end of struct
* according to marker, write what was eaten in eaten_ in case of success
*/
bool gotoNextToken(std::istream & is, Marker & marker, 
  std::string &eaten_, const ParseOptions & po)
{
  TRACE_ME;
  eaten_.clear();
  TRACE(3, "gotoNextToken with marker "<<marker);
  if (po.ignoreIndent)
    TRACE(3, "ignoring indentation");
  std::string spaceStr;
  std::string term;
  int c = EOF;
  //get spaces
  while (is.good())
  {
    c = is.get();
    TRACE(49, "char "<<c);
    if (!isspace(c))
    {
      break;
    }
    spaceStr += c;
  }
  if (!is.good())
  {
    TRACE(3, " FALSE, EOF");
    return false;
  }
  if (c == '}')
    term = "}";
  else if (c=='<')
  {
    term += c;
    while (is.good())
    {
      c = is.get();
      term += c;
      if (c == '>')
	break;
    }
    if (term.length() > 1 && term[1] != '/')
    {
      //not a closing tag
      undoStream(is, term);
      term.clear();
    }
  }
  else
    undoStream(is, c);
  
  bool newLine = false;
  int spaces = 0;
  size_t p = spaceStr.find_last_of('\n');
  if (p != spaceStr.npos)
  {
    newLine = true;
    spaces = spaceStr.length() - p - 1;
  }
  TRACE(6, "got marker [" << newLine <<',' << spaces << ']');
  if (!term.empty())
  {
    if (marker.terminator == term)
    {
      TRACE(3, " FALSE, found terminator");
      return false; //eat it
    }
    else
    {
      if (!marker.terminator.empty())
	PARSE_ERROR(is, "expected "<<marker.terminator);
      TRACE(3, " FALSE, parent terminator");
      undoStream(is, term);
      undoStream(is, spaceStr);
      return false;
    }
  }
  
  //XXX we chose to be nice and ignore newline if terminator expected
  if (!marker.terminator.empty())
  {
    TRACE(3," TRUE, terminator not present");
    eaten_ = spaceStr; //spaces+term but term is empty
    return true;
  }
  if (!marker.enforceNewLine)
  {
    TRACE(3, "  newline autodetection triggered:" << newLine);
    marker.enforceNewLine = true;
    marker.newLine = newLine;
  }
  if (!marker.enforceIndent)
  {
    TRACE(3, "  indent autodetection triggered:" << spaces);
    marker.enforceIndent = true;
    marker.spaces = spaces;
  }
  if (newLine && !marker.newLine)
  {
    TRACE(3, " FALSE, unexpected newline");
    undoStream(is, spaceStr);
    return false;
  }
  if (!newLine && marker.newLine)
    PARSE_ERROR(is, "expected newline not found");
  if (!po.ignoreIndent && spaces > marker.spaces)
  {
    //would be fun, but not working:
    //accept but leave the extra spaces here
    /*TRACE(3, " TRUE, all ok, leaving extra spaces");
    eaten_ = spaceStr.substr(0, spaceStr.length()-(spaces-marker.spaces));
    undoStream(is, spaceStr.substr(spaceStr.length()-(spaces-marker.spaces), 
      spaceStr.npos));
    return true;*/
    PARSE_ERROR(is, "indentation error: expected "<<marker.spaces<<" got "
      <<spaces);
  }
  /* There is a dumb case where this incorectly trigger with marker.newLine
  unset: containers. It is dumb to count spaces if !newLine.
  */
  if (!po.ignoreIndent && spaces < marker.spaces && marker.newLine)
  {
    TRACE(3, " FALSE, not enough spaces");
    undoStream(is, spaceStr);
    return false; 
  }
  eaten_ = spaceStr;
  eaten_ += term;
  TRACE(3, " TRUE, all ok, eaten_='" << eaten_ << "'");
  return true;
}

class Empty 
{
public:
  Empty():set(false) {}
  bool isSet() {return set;}
  bool set;
};

std::istream & operator >> (std::istream &i, Empty &e)
{
  e.set = true;
  return i;
}

std::ostream & operator << (std::ostream& o, const Empty &)
{
  return o;
}

class Flag
{
public:
  OptionState state;
  Flag() :state(UNSET) {}
  bool isTrue() {return state != DENY && state != UNSET;}
  bool isFalse() {return state == DENY;}
};

std::istream & operator >> (std::istream &i, Flag &f)
{
  Token s;
  i >> s;
  TRACE(45, "parsing flag with '"<<s<<"'");
  if (!strcasecmp(s.c_str(), "allow")
    ||  !strcasecmp(s.c_str(), "allowed")
    )
    f.state = ALLOW;
  else if (!strcasecmp(s.c_str(), "deny") 
    || !strcasecmp(s.c_str(), "denied") 
    || !strcasecmp(s.c_str(), "no"))
    f.state = DENY;
  else if (!strcasecmp(s.c_str(), "require") 
        || !strcasecmp(s.c_str(), "required")
        ||  !strcasecmp(s.c_str(), "yes"))
    f.state = REQUIRE;
  else
  {
    f.state = REQUIRE;
    undoStream(i, s);
  }
  return i;
}
std::ostream & operator << (std::ostream& o, const Flag &)
{
  return o;
}


//TODO: explain this!!
#define FIELD_OPTIONSTATE(f) Flag f; Empty allow##f; Empty require##f;\
   Empty deny##f; Empty no##f;
#define IS_ALLOW(f) (f.state == ALLOW || allow##f.isSet())
#define IS_DENY(f) (f.state == DENY || deny##f.isSet() || no##f.isSet())
#define IS_REQUIRE(f) (f.state == REQUIRE || require##f.isSet())
#define FIELD_BOOL(f) Flag f; Empty not##f; Empty no##f;Empty enable##f;\
  Empty disable##f; Empty allow##f; Empty deny##f;

//warning IS_TRUE is not equivalent to !IS_FALSE
#define IS_TRUE(f) (f.isTrue() || enable##f.isSet() || allow##f.isSet())
#define IS_FALSE(f) (f.isFalse() || not##f.isSet() || disable##f.isSet()  \
  ||deny##f.isSet() || no##f.isSet())

#define BOOL_WRITE(b, f) if IS_TRUE(f) b=true; if (IS_FALSE(f)) b = false;
#define OPTIONSTATE_WRITE(b, f) if (IS_ALLOW(f)) b = ALLOW; if (IS_DENY(f)) \
  b = DENY; if (IS_REQUIRE(f)) b = REQUIRE;
#define CHECK_BOOL

//I don't dare to use BOOST_PP here, doxygen needs to parse this TODO: try
class LeniantParseOptions{
public:
  FIELD_BOOL(optional)
  FIELD_BOOL(multipleWrite)
  FIELD_BOOL(ignoreIndent)
  FIELD_OPTIONSTATE(fieldName);
  FIELD_OPTIONSTATE(terminator);
  FIELD_OPTIONSTATE(singleLine);
  FIELD_OPTIONSTATE(colonAfterFieldName);
  FIELD_OPTIONSTATE(equalAfterFieldName);
  FIELD_OPTIONSTATE(xmlFieldName);
  FIELD_OPTIONSTATE(dashInList);
  void fill(ParseOptions &po)
  {
    //fill all fields, leaving them alone if nothing is said in conf
    BOOL_WRITE(po.optional, optional)
    BOOL_WRITE(po.multipleWrite, multipleWrite)
    BOOL_WRITE(po.ignoreIndent, ignoreIndent)
    OPTIONSTATE_WRITE(po.fieldName, fieldName)
    OPTIONSTATE_WRITE(po.terminator, terminator)
    OPTIONSTATE_WRITE(po.singleLine, singleLine)
    OPTIONSTATE_WRITE(po.colonAfterFieldName, colonAfterFieldName)
    OPTIONSTATE_WRITE(po.equalAfterFieldName, equalAfterFieldName)
    OPTIONSTATE_WRITE(po.xmlFieldName, xmlFieldName)
    OPTIONSTATE_WRITE(po.dashInList, dashInList)
  }
};

//we are in the namespace
#undef REGISTER_DESCTRAIT
#define REGISTER_DESCTRAIT(t, mt) \
template<> class DescTrait<t> {public: typedef mt type;}
#include "fsfc_leniantparseoptions.hh"
//RIGHTSHIFT_OPERATOR(LeniantParseOptions)
void ParseOptions::parse(const std::string &str)
{
  TRACE_ME;
  TRACE(10, "comment: '" <<str<<"'");
  ParseOptions po;
  po.allowNoise = true;
  LeniantParseOptions lpo;
  fsfcLeniantParseOptionsDesc::classOptions = po;
  std::istringstream ss;
  ss.str(str);
  try 
  {
    ss >> lpo;
  }
  catch(...)
  {
  }
  lpo.fill(*this);
  TRACE(10, "parse gave "<<*this);
}


void ParseOptions::merge(const ParseOptions &po)
{
  #define MERGE_BOOL(r, sname, fname)\
    if (sname.fname) fname = true;
  #define MERGE_OS(r, sname, fname) \
    if (sname.fname!= ALLOW) fname = sname.fname;
  BOOST_PP_LIST_FOR_EACH(MERGE_BOOL, po, PARSE_OPTION_BOOL_LIST)
  BOOST_PP_LIST_FOR_EACH(MERGE_OS, po, PARSE_OPTION_OPTIONSTATE_LIST)
  #undef MERGE_BOOL
  #undef MERGE_OS
}



void MetaString::fillStruct(std::istream &is, Marker &)
{
  int c;
  do 
  {
    c = is.get();
  } while (isspace(c) && is.good());
  if (c != '"')
  {
    undoStream(is, static_cast<char>(c));
    is >> ptr;
  }
  else
  { //read and unescape
    bool escape = false;
    while (is.good())
    {
      int c = is.get();
      if (!escape)
      {
	if (c == '\\')
	{
	  escape = true;
	  continue;
	}
	if (c == '"')
	  return;
	ptr += static_cast<char>(c);
      }
      else
      {
	switch(c)
	{
	case 'n':
	  ptr += '\n';
	  break;
	case 'r':
	  ptr += '\r';
	  break;
	case 't':
	  ptr += '\t';
	  break;
	case '\\':
	case '"':
	default:
	  ptr += static_cast<char>(c);
	}
	escape = false;
      }
    }
  }
}


using std::string;
REGISTER_NAME(string, string, MetaString);
REGISTER_NAME(std::string, stdstring, MetaString);

}
